# do not change this line manually
# this is managed by git tag and updated on every release
__version__ = '0.0.47'

# do not change this line manually
# this is managed by shell/make-proto.sh and updated on every execution
__proto_version__ = '0.0.10'
